/*
 * Translator.java
 * Copyright (C) 2004-2005, Bjorn Bringert (bringert@cs.chalmers.se)
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package se.chalmers.cs.gf.translate;

import se.chalmers.cs.gf.GFException;
import se.chalmers.cs.gf.abssyn.*;
import se.chalmers.cs.gf.parse.*;
import se.chalmers.cs.gf.linearize.*;
import se.chalmers.cs.gf.linearize.gfvalue.*;
import se.chalmers.cs.gf.util.Pair;
import static se.chalmers.cs.gf.util.Pair.pair;

import java.util.*;

/**
 *  Translates between different concrete syntaces which use the
 *  same abstract syntax.
 */
public class Translator {

        private String name;

        private Parsers parsers;

        private TypeAnnotator annotator;

        private Linearizers linearizers;

        /**
         *  @param name The name to use for the translator.
         *  @param parsers The parsers that this translator should use.
         *  @param annotator A type annotator.
         *  @param linearizers The linearizers that this translator should use.
         */
        public Translator(String name, Parsers parsers, 
                          TypeAnnotator annotator, Linearizers linearizers) {
                this.name = name;
                this.parsers = parsers;
                this.annotator = annotator;
                this.linearizers = linearizers;
        }

        /**
         *  Gets the name of this translator.
         */
        public String getName() {
                return name;
        }

        /**
         *  Parse a string as a given language, using the default
	 *  start category for that language.
         */
        public Collection<Tree> parse(String lang, String text) {
                Parser p = getParser(lang);
                return p.parse(text);
        }

        /**
         *  Parse a string as a given language, in the given category.
         */
        public Collection<Tree> parse(String lang, String text, String cat) {
                Parser p = getParser(lang);
                return p.parse(text,cat);
        }

        /**
         *  Try parsing a string with all the available parsers.
         */
        public Collection<Pair<String,Tree>> parseWithAll(String text) {
                return ParseAll.parseWithAll(parsers, text);
        }

        /**
         *  Does type annotation and linearization.
         */
        public String linearize(String lang, Tree tree) {
                Tree annTree = annotator.annotate(tree);
                Linearizer l = getLinearizer(lang);
                return valueAsString(l.linearize(annTree));
        }

        /**
         *  Does type annotation and linearization with all available linearizers.
         */
        public Collection<Pair<String,String>> linearizeWithAll(Tree tree) {
                Tree annTree = annotator.annotate(tree);
                List<Pair<String,Value>> ps = LinearizeAll.linearizeWithAll(linearizers, annTree);
                List<Pair<String,String>> ret = new ArrayList<Pair<String,String>>(ps.size());
                for (Pair<String,Value> p : ps)
                        ret.add(pair(p.fst, valueAsString(p.snd)));
                return ret;
        }

        /**
         *  Translate a string from one language to another.
         */
        public Set<String> translate(String fromLang, String toLang, String text) {
                Collection<Tree> trees = parse(fromLang, text);

                Set<String> output = new HashSet<String>();
                for (Tree tree : trees) {
                        try {
                                output.add(linearize(toLang, tree));
                        } catch (GFException ex) {
                                ex.printStackTrace();
                        }
                }

                return output;
        }

        /**
         *  List all the available input languages.
         */
        public Set<String> listParsers() {
                return parsers.listParsers();
        }

        /**
         *  List all the available output languages.
         */
        public Set<String> listLinearizers() {
                return linearizers.listLinearizers();
        }

        /**
         *  Get the parser for the given language.
         */
        public Parser getParser(String lang) {
                return parsers.getParser(lang);
        }

        /**
         *  Get the linearizer for the given language.
         */
        public Linearizer getLinearizer(String lang) {
                return linearizers.getLinearizer(lang);
        }

        private String valueAsString(Value value) {
                 Record r = (Record)value;
                 Value v = r.get("s"); // FIXME: is there always an s field?
                 String s = v.toString(); // FIXME: pick the first form if it is a table
                 return s;
        }


}
